@@ -134,7 +134,7 @@ class Agent(object):
 			#self.initialize()
 			if smart_card:
 				path = smart_cards.path
-				if OS in ["osx"]:
+				if OS in ["macos"]:
 					os.system(f"rm -fr {smart_cards.path}")
 					os.system(f"cp {smart_cards.original_path} {smart_cards.path}")
 					os.system(f"chmod 644 {smart_cards.path}")
