@@ -27,7 +27,7 @@ try:
 	import os, sys, requests, ast, json, pathlib, glob, platform, subprocess, pexpect, random, getpass, time
 
 	# inc imports.
-	from fil3s import Files, Formats
+	from fil3s import Files, Formats, gfp
 	from r3sponse import r3sponse
 	import cl1, syst3m, encrypti0n, netw0rk
 
@@ -51,7 +51,8 @@ except ImportError as e:
 # source.	
 SOURCE_PATH = syst3m.defaults.source_path(__file__, back=3)
 BASE = syst3m.defaults.source_path(SOURCE_PATH)
-OS = syst3m.defaults.operating_system(supported=["linux", "osx"])
+OS = syst3m.defaults.operating_system(supported=["linux", "macos"])
+LOG_LEVEL = syst3m.defaults.log_level(default=0)
 syst3m.defaults.alias(alias=ALIAS, executable=f"{SOURCE_PATH}")
 
 # universal variables.
@@ -60,7 +61,7 @@ GROUP = "root"
 HOME_BASE = "/home/"
 HOME = f"/home/{USER}/"
 MEDIA = f"/media/{USER}/"
-if OS in ["osx"]: 
+if OS in ["macos"]: 
 	HOME_BASE = "/Users/"
 	HOME = f"/Users/{USER}/"
 	MEDIA = f"/Volumes/"
