@@ -33,7 +33,7 @@ class SmartCards(object):
 			#self.original_path = "/usr/local/lib/libykcs11.dylib"
 			#self.path = "/usr/local/lib/libykcs11_NOTALNK.dylib"
 			#self.original_path = "/usr/local/lib/opensc-pkcs11.so"
-			self.original_path = f"{SOURCE_PATH}/lib/opensc/osx/opensc-pkcs11.so"
+			self.original_path = f"{SOURCE_PATH}/lib/opensc/macos/opensc-pkcs11.so"
 			self.path = "/usr/local/lib/opensc-pkcs11_NOTALINK.so"
 			if not os.path.exists(self.original_path):
 				raise OSError("opensc package is not installed, run: [$ brew install yubico-piv-tool opensc].")
