@@ -6,12 +6,13 @@ from ssht00ls.classes.config import *
 from ssht00ls.classes import utils
 from ssht00ls.classes.aliases import aliases
 from ssht00ls.classes.agent import agent
+import ssht00ls.classes.ssync.utils as ssync_utils 
 
 # the ssync object class.
 class SSync(object):
 	def __init__(self):
-		self.exclude = ['__pycache__', '.DS_Store']
 		aliases.sync()
+		self.utils = ssync_utils
 		self.sync()
 		self.gfp = Formats.FilePath("global")
 	def mount(self, 
@@ -40,9 +41,10 @@ class SSync(object):
 			"forced:bool":forced,
 			"accept_new_host_keys:bool":accept_new_host_keys, })
 		if not response.success: return response
-		elif not forced and os.path.exists(path):
+		cache_path = gfp.clean(path.split(" (d)")[0], remove_last_slash=True)
+		if not forced and os.path.exists(path):
 			return r3sponse.error_response(f"Path [{path}] already exists.")
-		elif "*mounted*" in str(cache.get(id=path, group="daemons")):
+		elif "*mounted*" in str(cache.get(id=cache_path, group="daemons")):
 			return r3sponse.error_response(f"Path [{path}] is already mounted.")
 		path = self.gfp.clean(path=path)
 		remote = self.gfp.clean(path=remote)
@@ -60,6 +62,11 @@ class SSync(object):
 		if not response.success: return response
 
 		# start daemon.
+		cache.set(id=cache_path, group="daemons", data="*mounted*")
+		self.set_mounted_icon(path)
+		#status = cache.get(id=cache_path, group="daemons")
+		#if status != "*mounted*":
+		#	return r3sponse.error_response(f"Failed to set the {path} daemon status.")
 		daemon = self.daemon(alias=alias, path=path, remote=remote, start=True)
 
 		# handler.
@@ -83,15 +90,17 @@ class SSync(object):
 			"forced:bool":forced,
 			"sudo:bool":sudo, })
 		if not response.success: return response
-		path = utils.clean_path(path)
-		mounted = "*mounted*" in str(cache.get(id=path, group="daemons"))
+		path = gfp.clean(path)
+		cache_path = gfp.clean(path.split(" (d)")[0], remove_last_slash=True)
+		mounted = "*mounted*" in str(cache.get(id=cache_path, group="daemons"))
 		if not mounted:
 			if not os.path.exists(path):
 				return r3sponse.error_response(f"Path [{path}] does not exist.")
 			elif not os.path.isdir(path):
 				return r3sponse.error_response(f"Path [{path}] is not a directory.")
-			elif "*mounted*" not in str(cache.get(id=path, group="daemons")):
-				return r3sponse.error_response(f"Path [{path}] is not mounted.")
+			status = cache.get(id=cache_path, group="daemons")
+			if "*mounted*" not in str(status):
+				return r3sponse.error_response(f"Path [{path}] is not mounted (status: {status}).")
 
 		# wait for daemon stop.
 		response = self.stop_daemon(path=path)
@@ -106,7 +115,7 @@ class SSync(object):
 		# checks.
 		if path == None:
 			return r3sponse.error_response(f"Define parameter: path.")
-		path = utils.clean_path(path)
+		path = gfp.clean(path)
 
 		# remote.
 		if alias != None:
@@ -149,28 +158,10 @@ class SSync(object):
 				elif not os.path.isdir(path):
 					return r3sponse.error_response(f"Path [{path}] is not a directory.")
 
-			# index.
-			index, gfp, dir, ids = Files.Dictionary(path=False, dictionary={}), Formats.FilePath("/"), Files.Directory(path=path), []
-			for _path_ in dir.paths(recursive=True, files_only=True):
-				fp = Formats.FilePath(_path_)
-				if fp.name() not in [".DS_Store"] and fp.basename() not in ["__pycache__"]:
-					id = _path_
-					if fp.directory(): 
-						id += " (d)"
-						if os.listdir(_path_) == []: id += " (e)"
-					if id not in ids: ids.append(id)
-					index[id] = fp.mtime(format="seconds")
-			for _path_ in dir.paths(recursive=True, dirs_only=True):
-				id = _path_+" (d)"
-				if os.listdir(_path_) == []: id += " (e)"
-				if id not in ids: 
-					index[id] = gfp.mtime(path=_path_, format="seconds")
-					ids.append(id)
-			index.dictionary = index.sort(alphabetical=True)
-
 			# handler.
-			return r3sponse.success_response(f"Successfully indexed {len(index.dictionary)} files from directory [{path}].", {
-				"index":index.dictionary,
+			dict = self.utils.index(path)
+			return r3sponse.success_response(f"Successfully indexed {len(dict)} files from directory [{path}].", {
+				"index":dict,
 			})
 
 			#
@@ -185,7 +176,7 @@ class SSync(object):
 		# exlude subpaths (list) (leave None to use default).
 		exclude=None,
 		# path is directory boolean (leave None to parse automatically).
-		directory=None,
+		directory=True,
 		empty_directory=False,
 		# update deleted files.
 		delete=False,
@@ -199,79 +190,46 @@ class SSync(object):
 		checks=True,
 		# log level.
 		log_level=0,
+		# get the command in str.
+		command=False,
 	):	
-		# checks.
 		if checks:
-
-			# check alias.
-			path = utils.clean_path(path)
-			remote = utils.clean_path(remote)
-			response = aliases.check(alias)
-			if not response.success: return response
-			
-			# check passphrase.
 			if CONFIG["aliases"][alias]["smart_card"] in [True, "true", "True"]:
 				response = ENCRYPTION.decrypt(CONFIG["aliases"][alias]["passphrase"])
 			else:
 				response = ENCRYPTION.decrypt(CONFIG["aliases"][alias]["passphrase"])
 			if not response.success: return response
 			passphrase = response.decrypted.decode()
-			
-			# tests.
+			response = aliases.check(alias)
+			if not response.success: return response
 			response = agent.add(path=CONFIG["aliases"][alias]["private_key"], passphrase=passphrase)
 			if not response["success"]: return response
-			response = utils.test_ssht00ls(alias=alias, accept_new_host_keys=accept_new_host_keys)
-			if not response.success: return response
-			response = utils.test_ssh_path(alias=alias, path=remote, accept_new_host_keys=accept_new_host_keys)
-			if not response.success: return response
-
-		# dir.
-		if directory == None: 
-			response = utils.test_ssh_dir(alias=alias, path=remote, accept_new_host_keys=accept_new_host_keys)
-			if not response.success and "not a directory" not in response.error: return response
-			elif response.success:
-				directory = True
-			else: directory = False
-			tested = True
-		elif directory and checks:
-			response = utils.test_ssh_dir(alias=alias, path=remote, accept_new_host_keys=accept_new_host_keys)
-			if not response.success: return response
-			tested = True
-
-		# check base.
-		base = Formats.FilePath(path).base(back=1)
-		if not os.path.exists(base): 
-			os.system(f"mkdir -p {base}")
-			if not os.path.exists(base): 
-				return r3sponse.error_response(f"Failed to create pull base {base}.")
-			if log_level >= 3:
-				print(f"Created directory {base}.")
-
-		# fix rsync timestamp bug.
-		if empty_directory and directory and not os.path.exists(path):
-			os.system(f"mkdir -p {path}")
-
-		# pull.
-		if exclude == None: exclude = self.exclude
-		exclude_str = ""
-		for i in exclude: exclude_str += f"--exclude {i} "
-		delete_str = Formats.Boolean(delete).convert(true="--delete", false="")
-		command = f"rsync -{Formats.Boolean(directory).convert(true='a', false='')}zqt {alias}:{remote} {path} {exclude_str}{delete_str}"
-		loader = syst3m.console.Loader(f"Pulling [{alias}:{remote}] to [{path}]", interactive=False)
-		output = syst3m.utils.__execute_script__(command)
-		if len(output) > 0 and output[len(output)-1] == "\n": output = output[:-1]
-		cache.set(id=path, data="*mounted*", group="daemons")
-		
-		# handler.
-		if "rsync: " in output or "rsync error: " in output:
-			loader.stop(success=False)
-			print(output)
-			return r3sponse.error_response(f"Failed to pull [{alias}:{remote}] to [{path}].")
-		else:
-			loader.stop()
-			return r3sponse.success_response(f"Successfully pulled [{alias}:{remote}] to [{path}].")
-
-		#
+		return self.utils.pull(
+			# the local path.
+			path=path, 
+			# the ssht00ls alias.
+			alias=alias, 
+			# the remote path.
+			remote=remote, 
+			# exlude subpaths (list) (leave None to use default).
+			exclude=exclude,
+			# path is directory boolean (leave None to parse automatically).
+			directory=directory,
+			empty_directory=empty_directory,
+			# update deleted files.
+			delete=delete,
+			# forced mode.
+			forced=forced,
+			# version control.
+			safe=safe,
+			# accept new hosts keys.
+			accept_new_host_keys=accept_new_host_keys,
+			# checks.
+			checks=checks,
+			# log level.
+			log_level=log_level,
+			# get the command in str.
+			command=command,)
 	def push(self,
 		# the local path.
 		path=None, 
@@ -282,7 +240,7 @@ class SSync(object):
 		# exlude subpaths (list) (leave None to use default).
 		exclude=None,
 		# path is directory boolean (leave None to parse automatically).
-		directory=None,
+		directory=True,
 		empty_directory=False,
 		# update deleted files.
 		delete=False,
@@ -296,66 +254,46 @@ class SSync(object):
 		checks=True,
 		# log level.
 		log_level=0,
+		# get the command in str.
+		command=False,
 	):
-		# checks.
 		if checks:
-
-			# check alias.
-			path = utils.clean_path(path)
-			remote = utils.clean_path(remote)
-			response = aliases.check(alias)
-			if not response.success: return response
-			
-			# check passphrase.
 			if CONFIG["aliases"][alias]["smart_card"] in [True, "true", "True"]:
 				response = ENCRYPTION.decrypt(CONFIG["aliases"][alias]["passphrase"])
 			else:
 				response = ENCRYPTION.decrypt(CONFIG["aliases"][alias]["passphrase"])
 			if not response.success: return response
 			passphrase = response.decrypted.decode()
-			
-			# tests.
+			response = aliases.check(alias)
+			if not response.success: return response
 			response = agent.add(path=CONFIG["aliases"][alias]["private_key"], passphrase=passphrase)
 			if not response["success"]: return response
-			response = utils.test_ssht00ls(alias=alias, accept_new_host_keys=accept_new_host_keys)
-			if not response.success: return response
-			response = utils.test_ssh_path(alias=alias, path=Formats.FilePath(remote).base(), accept_new_host_keys=accept_new_host_keys)
-			if not response.success: return response
-
-		# dir.
-		if directory == None: directory = os.path.isdir(path)
-		elif directory and not os.path.isdir(path):
-			return r3sponse.error_response(f"Path {path} is not a directory.")
-
-		# check remote base.
-		base = Formats.FilePath(remote).base(back=1)
-		response = utils.test_ssh_dir(alias=alias, path=base, accept_new_host_keys=accept_new_host_keys, create=True, checks=False)
-		if not response.success: return response
-		if response.created and log_level >= 3:
-				print(f"Created remote directory {base}.")
-
-		# push.
-		if exclude == None: exclude = self.exclude
-		exclude_str = ""
-		for i in exclude: exclude_str += f"--exclude {i} "
-		delete_str = Formats.Boolean(delete).convert(true="--delete", false="")
-		command = f"rsync -{Formats.Boolean(directory).convert(true='a', false='')}zqt {path} {alias}:{remote} {exclude_str}{delete_str}"
-		loader = syst3m.console.Loader(f"Pushing [{path}] to [{alias}:{remote}].", interactive=False)
-		output = syst3m.utils.__execute_script__(command)
-		if len(output) > 0 and output[len(output)-1] == "\n": output = output[:-1]
-		cache.set(id=path, data="*mounted*", group="daemons")
-		
-		# handler.
-		if "rsync: " in output or "rsync error: " in output:
-			loader.stop(success=False)
-			print(output)
-			return r3sponse.error_response(f"Failed to push [{path}] to [{alias}:{remote}].")
-		else:
-			loader.stop()
-			return r3sponse.success_response(f"Successfully pushed [{path}] to [{alias}:{remote}].")
-
-
-		#
+		return self.utils.push(
+			# the local path.
+			path=path, 
+			# the ssht00ls alias.
+			alias=alias, 
+			# the remote path.
+			remote=remote, 
+			# exlude subpaths (list) (leave None to use default).
+			exclude=exclude,
+			# path is directory boolean (leave None to parse automatically).
+			directory=directory,
+			empty_directory=empty_directory,
+			# update deleted files.
+			delete=delete,
+			# forced mode.
+			forced=forced,
+			# version control.
+			safe=safe,
+			# accept new hosts keys.
+			accept_new_host_keys=accept_new_host_keys,
+			# checks.
+			checks=checks,
+			# log level.
+			log_level=log_level,
+			# get the command in str.
+			command=command,)
 	# the mounted daemon.
 	def sync(self):
 		for path in Files.Directory(path=f"{cache.path}/daemons/").paths(recursive=False):
@@ -370,7 +308,7 @@ class SSync(object):
 		# settings.
 		start=True,
 		# the daemons log level.
-		log_level=3,
+		log_level=LOG_LEVEL,
 		# sandbox (do not delete any files).
 		sandbox=False,
 		# serialized.
@@ -382,11 +320,11 @@ class SSync(object):
 				"remote":None,
 				"path":None,
 				"start":True,
-				"log_level":3,
+				"log_level":LOG_LEVEL,
 				"sandbox":False,
 			})
-		path = utils.clean_path(path)
-		remote = utils.clean_path(remote)
+		path = gfp.clean(path)
+		remote = gfp.clean(remote)
 		daemon = self.Daemon({
 			"alias":alias,
 			"remote":remote,
@@ -398,11 +336,34 @@ class SSync(object):
 		})
 		if start: webserver.start_daemon(daemon, group="daemons", id=daemon.id)
 		return daemon
+	def set_mounted_icon(self, path):
+		if OS in ["macos", "macos"]:
+			#os.system(f"cp '{SOURCE_PATH}/static/media/icons/Icon\r' '{path}/Icon\r'")
+			icon = f'{SOURCE_PATH}/static/media/icons/mounted.png'
+			syst3m.utils.__execute_script__(f"""
+				
+				# Take an image and make the image its own icon:
+				sips -i {icon}
+
+				# Extract the icon to its own resource file:
+				/Developer/Tools/DeRez -only icns {icon} > /tmp/tmpicns.rsrc
+
+				# append this resource to the file you want to icon-ize.
+				/Developer/Tools/Rez -append /tmp/tmpicns.rsrc -o {path}
+
+				# Use the resource to set the icon.
+				/Developer/Tools/SetFile -a C {path}
+
+				# clean up.
+				rm  -fr /tmp/tmpicns.rsrc
+
+				""")
 	def stop_daemon(self, path, timeout=60, sleeptime=1):
-		cache.set(id=path, data="*stop*", group="daemons")
+		cache_path = gfp.clean(path.split(" (d)")[0], remove_last_slash=True)
+		cache.set(id=cache_path, data="*stop*", group="daemons")
 		stopped = False
 		for i in range(int(timeout/sleeptime)):
-			status_ = cache.get(id=path, group="daemons")
+			status_ = cache.get(id=cache_path, group="daemons")
 			if "*stopped*" in status_ or "*crashed*" in status_:
 				stopped = True
 				break
@@ -413,14 +374,16 @@ class SSync(object):
 			return r3sponse.error_response(f"Failed to stop ssht00ls daemon [{path}].")
 	class Daemon(syst3m.objects.Thread):
 		def __init__(self, attributes={}):
+			self.utils = ssync_utils
 			syst3m.objects.Thread.__init__(self)
 			self.assign(attributes)
-			self.path = utils.clean_path(self.path)
+			self.path = gfp.clean(self.path)
 			self.id = f"{self.alias}:{self.remote} {self.path}"
 			try:self.log_level
 			except: self.log_level = -1
 			self.last_index = {}
 			self.last_remote_index = {}
+			self.cache_path = gfp.clean(self.path.split(" (d)")[0], remove_last_slash=True)
 		def run(self):
 
 			# logs.
@@ -428,9 +391,10 @@ class SSync(object):
 				loader = syst3m.console.Loader(f"Checking daemon {self.id}", interactive=INTERACTIVE)
 
 			# checks.
-			if "*mounted*" not in str(cache.get(id=self.path, group="daemons")):
+			status = cache.get(id=self.cache_path, group="daemons")
+			if "*mounted*" not in str(status):
 				if self.log_level >= 0: loader.stop(success=False)
-				self.crash(f"ssht00ls daemon ({self.id}): Path [{self.path}] is not mounted.")
+				self.crash(f"ssht00ls daemon ({self.id}): Path [{self.path}] is not mounted (status: {status}).")
 			if self.log_level >= 0: 
 				loader.mark(f"Checking mountpoint {self.path}")
 			if not os.path.exists(self.path):
@@ -466,7 +430,7 @@ class SSync(object):
 			while True:
 
 				# check stop command.
-				status = cache.get(id=self.path, group="daemons")
+				status = cache.get(id=self.cache_path, group="daemons")
 				if "*stop*" in status or "*unmounting*" in status:
 					break
 
@@ -491,22 +455,22 @@ class SSync(object):
 		def stop():
 			response = self.unmount()
 			if not response["success"]: self.crash(f"ssht00ls daemon ({self.id}) error: {response['error']}")
-			cache.set(id=self.path, data="*stopped*", group="daemons")
+			cache.set(id=self.cache_path, data="*stopped*", group="daemons")
 			if self.log_level >= 0: print(f"Stopped daemon {self.id}")
 		def crash(self, error=None, response=None, unmount=True):
+			if response != None: error = response["error"]
 			if unmount:
 				response = self.unmount()
-				if not response["success"]: 
-					cache.set(id=self.path, data="*crashed*", group="daemons")
+				if not response.success: 
+					cache.set(id=self.cache_path, data="*crashed*", group="daemons")
 					raise ValueError(f"ssht00ls daemon ({self.id}) error: {response['error']}")
-			if response != None: error = response["error"]
-			cache.set(id=self.path, data="*crashed*", group="daemons")
+			cache.set(id=self.cache_path, data="*crashed*", group="daemons")
 			raise ValueError(error)
 		def unmount(self):
 			if os.path.exists(self.path):
 				response = self.sync()
 				if not response.success: return response
-				cache.set(id=self.path, data="*unmounting*", group="daemons")
+				cache.set(id=self.cache_path, data="*unmounting*", group="daemons")
 				time.sleep(1)
 				response = self.delete(self.path, remote=False, subpath=False,)
 				if not response.success: return response
@@ -545,7 +509,7 @@ class SSync(object):
 					"remote_index":remote_index,
 				})
 			else: return self.process_index(clean_index, remote_clean_index, mismatches)
-
+			#
 		def process_index(self, index, remote_index, mismatches):
 
 			# iterate.
@@ -627,39 +591,47 @@ class SSync(object):
 						self.set_last_index(index) ; self.set_last_index(remote_index, remote=True)
 						return r3sponse.error_response(f"Unable to compare rmtime: {rmtime} & lmtime: {lmtime}. (path: {subpath}), (rmtime: {rmtime}), (lmtime: {lmtime}), (last_rmtime: {last_rmtime}), (last_lmtime: {last_lmtime}), index: {json.dumps(index,indent=4)}, remote index: {json.dumps(remote_index,indent=4)}.")
 
-				# should not happen.
+				# exceptions.
 				else:
 					self.set_last_index(index) ; self.set_last_index(remote_index, remote=True)
-					return r3sponse.error_response(f"Should not happen (#27834628). (path: {subpath}), (rmtime: {rmtime}), (lmtime: {lmtime}), (last_rmtime: {last_rmtime}), (last_lmtime: {last_lmtime}), index: {json.dumps(index,indent=4)}, remote index: {json.dumps(remote_index,indent=4)}.")
+					return r3sponse.error_response(f"Should not happen (#407294). (path: {subpath}), (rmtime: {rmtime}), (lmtime: {lmtime}), (last_rmtime: {last_rmtime}), (last_lmtime: {last_lmtime}), index: {json.dumps(index,indent=4)}, remote index: {json.dumps(remote_index,indent=4)}.")
 
-				# add to updates.
-				if local_to_remote and remote_to_local:
-					self.set_last_index(index) ; self.set_last_index(remote_index, remote=True)
-					return r3sponse.error_response(f"Can not synchronize both remote to local & local to remote (rmtime: {rmtime}) (lmtime: {lmtime}), (last_rmtime: {last_rmtime}) (last_lmtime: {last_lmtime}), index: {json.dumps(index,indent=4)}, remote index: {json.dumps(remote_index,indent=4)}.")
-				if "delete" in options: 
-					if self.log_level >= 1: 
-						if local_to_remote:
-							print(f"Deletion required {lfullpath} {self.alias}:{rfullpath} (rmtime: {rmtime}) (lmtime: {lmtime})")
-						else:
-							print(f"Deletion required {self.alias}:{rfullpath} {lfullpath} (rmtime: {rmtime}) (lmtime: {lmtime})")
-					deletions[subpath] = {
-						"options":options,
-						"remote_to_local":remote_to_local,
-						"local_to_remote":local_to_remote,
-						"directory":directory,
-						"empty_directory":empty_directory,}
-				elif remote_to_local or local_to_remote:
-					if self.log_level >= 1: 
-						if local_to_remote:
-							print(f"Update required {lfullpath} {self.alias}:{rfullpath} (rmtime: {rmtime}) (lmtime: {lmtime})")
-						else:
-							print(f"Update required {self.alias}:{rfullpath} {lfullpath} (rmtime: {rmtime}) (lmtime: {lmtime})")
-					updates[subpath] = {
-						"options":options,
-						"remote_to_local":remote_to_local,
-						"local_to_remote":local_to_remote,
-						"directory":directory,
-						"empty_directory":empty_directory,}
+				# do not remove this exception.
+				# it is required for safe edits, to make sure a dir never gets synced unless it is created / removed.
+				# also required by the multiprocessing idexing.
+				if directory and local_to_remote != None and remote_to_local != None:
+					pass
+					#return r3sponse.error_response(f"Should not happen (#143765). (path: {subpath}), (rmtime: {rmtime}), (lmtime: {lmtime}), (last_rmtime: {last_rmtime}), (last_lmtime: {last_lmtime}), index: {json.dumps(index,indent=4)}, remote index: {json.dumps(remote_index,indent=4)}.")
+				else:
+
+					# add to updates.
+					if local_to_remote and remote_to_local:
+						self.set_last_index(index) ; self.set_last_index(remote_index, remote=True)
+						return r3sponse.error_response(f"Can not synchronize both remote to local & local to remote (rmtime: {rmtime}) (lmtime: {lmtime}), (last_rmtime: {last_rmtime}) (last_lmtime: {last_lmtime}), index: {json.dumps(index,indent=4)}, remote index: {json.dumps(remote_index,indent=4)}.")
+					if "delete" in options: 
+						if self.log_level >= 1: 
+							if local_to_remote:
+								print(f"Deletion required {lfullpath} {self.alias}:{rfullpath} (rmtime: {rmtime}) (lmtime: {lmtime})")
+							else:
+								print(f"Deletion required {self.alias}:{rfullpath} {lfullpath} (rmtime: {rmtime}) (lmtime: {lmtime})")
+						deletions[subpath] = {
+							"options":options,
+							"remote_to_local":remote_to_local,
+							"local_to_remote":local_to_remote,
+							"directory":directory,
+							"empty_directory":empty_directory,}
+					elif remote_to_local or local_to_remote:
+						if self.log_level >= 1: 
+							if local_to_remote:
+								print(f"Update required {lfullpath} {self.alias}:{rfullpath} (rmtime: {rmtime}) (lmtime: {lmtime})")
+							else:
+								print(f"Update required {self.alias}:{rfullpath} {lfullpath} (rmtime: {rmtime}) (lmtime: {lmtime})")
+						updates[subpath] = {
+							"options":options,
+							"remote_to_local":remote_to_local,
+							"local_to_remote":local_to_remote,
+							"directory":directory,
+							"empty_directory":empty_directory,}
 			# handler.
 			self.set_last_index(index) ; self.set_last_index(remote_index, remote=True)
 			return r3sponse.success_response(f"Successfully indexed [{self.alias}:{rfullpath}] & [{lfullpath}].", {
@@ -669,7 +641,7 @@ class SSync(object):
 				"index":index,
 				"remote_index":remote_index,
 			})
-		def local_to_remote(self, path, info, directory=False, empty_directory=False, forced=False, delete=False):
+		def local_to_remote(self, path, info, directory=False, empty_directory=False, forced=False, delete=False, command=False):
 			if self.log_level >= 0: print(f"Synchronizing {path} to {self.alias}:{path} (delete: {delete}) (forced: {forced}).")
 			return self.ssync.push(
 				path=self.fullpath(path, append_slash=directory), 
@@ -682,8 +654,9 @@ class SSync(object):
 				safe=False,
 				accept_new_host_keys=True,
 				checks=False,
+				command=command,
 				log_level=self.log_level,)
-		def remote_to_local(self, path, info, directory=False, empty_directory=False, forced=False, delete=False):
+		def remote_to_local(self, path, info, directory=False, empty_directory=False, forced=False, delete=False, command=False):
 			if self.log_level >= 0: print(f"Synchronizing {self.alias}:{path} to {path} (delete: {delete}) (forced: {forced}).")
 			return self.ssync.pull(
 				path=self.fullpath(path, append_slash=directory), 
@@ -696,6 +669,7 @@ class SSync(object):
 				safe=False,
 				accept_new_host_keys=True,
 				checks=False,
+				command=command,
 				log_level=self.log_level,)
 		def synchronized(self, clean_index, remote_clean_index, all_paths):
 			mismatches = {}
@@ -741,6 +715,7 @@ class SSync(object):
 		def delete(self, path, remote=False, subpath=True):
 
 			# sandbox.
+			path = self.utils.serialize_path(gfp.clean(path))
 			if self.sandbox: 
 				if remote:
 					return r3sponse.success_response(f"Sandbox enabled, skip delettion of {self.alias}:{path}.")
@@ -758,7 +733,7 @@ class SSync(object):
 			
 			# local to remote.
 			if remote:
-				os.system(f"ssh {self.alias} ' printf 'y' | rm -fr {path} ' ")
+				os.system(f"""ssh {self.alias} " printf 'y' | rm -fr '{path}' " """)
 				response = utils.test_ssh_path(alias=self.alias, path=path)
 				if response.error != None and f"{path} does not exist" not in response.error:
 					return r3sponse.error_response(f"Failed to delete {path}, error: {response.error}")
@@ -766,11 +741,11 @@ class SSync(object):
 
 			# remote to local.
 			else:
-				os.system(f"printf 'y' | rm -fr {path}")
+				os.system(f"printf 'y' | rm -fr '{path}' ")
 				if os.path.exists(path):
 					return r3sponse.error_response(f"Failed to delete {path}")
 				return r3sponse.success_response(f"Successfully deleted {path}")
-		def sync(self):
+		def sync(self, multiprocessing=True, max_batch_count=50):
 
 			# get index.
 			response = self.index(short="auto")
@@ -782,34 +757,135 @@ class SSync(object):
 					return r3sponse.success_response(f"Directories [{self.alias}:{self.path}] & [{self.path}] are already synchronized.")
 			updates,deletions = response.unpack(["updates", "deletions"])
 
-			# iterate updates.
-			dirs = {}
-			for path, info in updates.items():
-				if self.log_level >= 0: print(f"Updating {path}.")
+			# parallel multiprocessing.
+			if multiprocessing:
+			
+				# order updates by dir depth from deepest to min.
+				# create depth index.
+				dir_updates, depths, max_depth = {}, {}, 0
+				for path, info in updates.items():
+					if info["directory"]: dir_updates[path] = info
+					else:
+						depth = len(gfp.clean(path, remove_double_slash=True, remove_first_slash=True, remove_last_slash=True).split("/"))
+						if depth >= max_depth: max_depth = depth
+						try: depths[str(depth)]
+						except KeyError: depths[str(depth)] = {}
+						depths[str(depth)][path] = info
+				
+				# add batches from highest depth to lowest.
+				update_batches, batch_count, inside_batch_count, last_depth = {}, 0, 0, None
+				for i in range(max_depth+1):
+					depth, found = max_depth-i, True
+					try: depths[str(depth)]
+					except KeyError: found = False
+					if found:
+						if (last_depth != None and last_depth != depth) or (inside_batch_count >= max_batch_count): 
+							last_depth = depth
+							batch_count += 1
+							inside_batch_count = 0
+						for path, info in depths[str(depth)].items():
+							if inside_batch_count >= max_batch_count: 
+								last_depth = depth
+								batch_count += 1
+								inside_batch_count = 0
+							try: update_batches[str(batch_count)]
+							except KeyError: update_batches[str(batch_count)] = {}
+							update_batches[str(batch_count)][path] = info
+							inside_batch_count += 1
+				# add directories.
+				for path, info in dir_updates.items():
+					if inside_batch_count >= max_batch_count: 
+						batch_count += 1
+						inside_batch_count = 0
+					try: update_batches[str(batch_count)]
+					except KeyError: update_batches[str(batch_count)] = {}
+					update_batches[str(batch_count)][path] = info
+					inside_batch_count += 1
 
-				# local to remote.
-				if info["local_to_remote"]:
-					if info["directory"] in [True, "True", "TRUE", "true"]: dirs[path] = info
-					else: self.local_to_remote(path, info)
+				# iterate updates.
+				commands = {}
+				for batch_count, batch in update_batches.items():
+					commands[batch_count] = []
+					for path, info in batch.items():
+						if self.log_level >= 0: print(f"Updating {path}.")
 
-				# remote to local.
-				elif info["remote_to_local"]:
-					#response = utils.test_ssh_dir(alias=self.alias, path=path)
-					#if response.error != None and "is not a directory" in response.error: return response
-					#elif response.success: dirs[path] = info
-					if info["directory"] in [True, "True", "TRUE", "true"]: dirs[path] = info
-					else: self.remote_to_local(path, info)
+						# local to remote.
+						if info["local_to_remote"]:
+							if info["directory"] in [True, "True", "TRUE", "true"]: 
+								commands[batch_count] += [self.local_to_remote(path, info, directory=True, command=True)]
+							else: 
+								commands[batch_count] += [self.local_to_remote(path, info, command=True)]
 
-			# iterate excepted dirs.
-			for path, info in dirs.items():
+						# remote to local.
+						elif info["remote_to_local"]:
+							if info["directory"] in [True, "True", "TRUE", "true"]: 
+								commands[batch_count] += [self.remote_to_local(path, info, directory=True, command=True)]
+							else: 
+								commands[batch_count] += [self.remote_to_local(path, info, command=True)]
 
-				# local to remote.
-				if info["local_to_remote"]:
-					self.local_to_remote(path, info, directory=True)
+				# execute commands.
+				for batch_count, batch in commands.items():
+					loader, c = None, len(batch)
+					if self.log_level >= 0: loader = f"Synchronizing {c} files."
+					msg = f"Successfully synchronized {c} files."
+					error = f"Failed to synchronize {c} files."
+					command = Files.Array(array=batch).string(joiner=" & ")
+					response = self.utils.execute(command=command, error=error, message=msg, loader=loader)
+					if not response.success:
+						response = self.utils.execute(command=command, error=error, message=msg, loader=loader)
+						if not response.success:
+							return response
+
+			# no multiprocessing.
+			else:
+
+				# iterate updates.
+				dirs, commands, c = {}, [], 0
+				for path, info in updates.items():
+					if self.log_level >= 0: print(f"Updating {path}.")
+
+					# local to remote.
+					if info["local_to_remote"]:
+						if info["directory"] in [True, "True", "TRUE", "true"]: dirs[path] = info
+						else: 
+							commands += [self.local_to_remote(path, info, command=True)]
+							c += 1
+
+					# remote to local.
+					elif info["remote_to_local"]:
+						#response = utils.test_ssh_dir(alias=self.alias, path=path)
+						#if response.error != None and "is not a directory" in response.error: return response
+						#elif response.success: dirs[path] = info
+						if info["directory"] in [True, "True", "TRUE", "true"]: dirs[path] = info
+						else: 
+							commands += [self.remote_to_local(path, info, command=True)]
+							c += 1
+
+				# iterate excepted dirs.
+				for path, info in dirs.items():
+
+					# local to remote.
+					if info["local_to_remote"]:
+						commands += [self.local_to_remote(path, info, directory=True, command=True)]
+						c += 1
+
+					# remote to local.
+					elif info["remote_to_local"]:
+						commands += [self.remote_to_local(path, info, directory=True, command=True)]
+						c += 1
+
+				# execute commands.
+				loader = None
+				if self.log_level >= 0: loader = f"Synchronizing {c} files."
+				msg = f"Successfully synchronized {c} files."
+				error = f"Failed to synchronize {c} files."
+				command = Files.Array(array=commands).string(joiner=" && ")
+				response = self.utils.execute(command=command, error=error, message=msg, loader=loader)
+				if not response.success:
+					response = self.utils.execute(command=command, error=error, message=msg, loader=loader)
+					if not response.success:
+						return response
 
-				# remote to local.
-				elif info["remote_to_local"]:
-					self.remote_to_local(path, info, directory=True)
 
 			# push deletions.
 			for path, info in deletions.items():
@@ -824,7 +900,7 @@ class SSync(object):
 			response = self.index(short=True)
 			if not response.success: return response
 			elif not response.synchronized:
-				if self.log_level <= 0:
+				if self.log_level <= -1:
 					return r3sponse.error_response(f"Failed to synchronize [{self.alias}:{self.remote}] & [{self.path}].")		
 				else:
 					index, remote_index = response.index, response.remote_index
@@ -904,6 +980,10 @@ class SSync(object):
 # initialized objects.
 ssync = SSync()
 
+#daemon = ssync.daemon(path="/tmp/testmount.small", remote="/tmp/testmount.small", alias="dev.vandenberghinc.com", start=False)
+#response = daemon.index(short=True)
+#print(response)
+#quit()
 
 #print(ssync.mount(
 #	alias="dev.vandenberghinc.com",
