@@ -23,7 +23,7 @@ def serialize_path(path, append_last_slash=False):
 	]:
 		path = path.replace(x,y)
 	while True:
-		if append_last_slash and len(path) > 0 and path[len(path)-1] == "/": path = path[:-1]
+		if append_last_slash and len(path) > 0 and path[len(path)-1] != "/": path = path+"/"
 		else: break
 	return path
 
@@ -68,7 +68,7 @@ def pull(
 	# checks.
 	checks=True,
 	# log level.
-	log_level=0,
+	log_level=LOG_LEVEL,
 	# get the command in str.
 	command=False,
 ):	
@@ -132,7 +132,8 @@ def pull(
 	lremote = serialize_path(gfp.clean(remote), append_last_slash=directory)
 	lpath = serialize_path(gfp.clean(path), append_last_slash=directory)
 	_command_ = f"rsync -{Formats.Boolean(directory).convert(true='a', false='')}zqt '{alias}:{lremote}' '{lpath}' {exclude_str} {include_str} {delete_str}"
-	if log_level >= 3: print(f"Rsync command: [{_command_}].")
+	#if log_level >= 3: 
+	print(f"Rsync command: [{_command_}].")
 
 	# execute.
 	if command: return _command_
@@ -172,7 +173,7 @@ def push(
 	# checks.
 	checks=True,
 	# log level.
-	log_level=0,
+	log_level=LOG_LEVEL,
 	# get the command in str.
 	command=False,
 ):
