@@ -205,7 +205,7 @@ def __set_file_path_ownership__(path, owner=os.environ.get("USER"), group=None,
 	if group == None: group = __get_empty_group__()
 	os.system(f"{sudo}chown {recursive}{owner}:{group} {path}")
 def __get_empty_group__():
-	if OS in ["osx"]: return "wheel"
+	if OS in ["macos"]: return "wheel"
 	elif OS in ["linux"]: return "root"
 def __delete_file_path__(path, sudo=False, forced=False):
 	if sudo: sudo = "sudo "
